#!/usr/bin/env python
#
#    WPSIG - WiFi Protected Setup Information Gathering
#    Copyright (C) 2010  Core Security Technologies
#
#    This file is part of WPSIG.
#
#    WPSIG is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    WPSIG is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with WPSIG.  If not, see <http://www.gnu.org/licenses/>.
#
#    Author: Andres Blanco 
#			 ablanco [at coresecurity.com]
#			 oss 	 [at coresecurity.com]

import re
import sys
import random
import struct
import datetime
import optparse

import pcapy
import PyLorcon2
from impacket import dot11
from impacket.dot11 import Dot11
from impacket.dot11 import Dot11Types
from impacket.dot11 import Dot11ManagementFrame
from impacket.dot11 import Dot11ManagementProbeRequest
from impacket.ImpactDecoder import RadioTapDecoder

WPS_DATA_ELEMENTS = {
                     0x1001 : "AP Channel", # 2 Bytes
                     0x1002 : "Association State", # 2 Bytes
                     0x1003 : "Authentication Type", # 2 Bytes
                     0x1004 : "Authentication Type Flags", # 2 Bytes
                     0x1005 : "Authenticator", # 8 Bytes
                     0x1008 : "Config Methods", # 2 Bytes
                     0x1009 : "Configuration Error", # 2 Bytes
                     0x100A : "Confirmation URL4", # <= 64 Bytes
                     0x100B : "Confirmation URL6", # <= 76 Bytes
                     0x100C : "Connection Type", # 1 Byte
                     0x100D : "Connection Type Flags", # 1 Byte
                     0x100E : "Credential",
                     0x1011 : "Device Name", # <= 32 Bytes
                     0x1012 : "Device Password ID", # 2 Bytes
                     0x1014 : "E-Hash1", # 32 Bytes
                     0x1015 : "E-Hash2", # 32 Bytes
                     0x1016 : "E-SNonce1", # 16 Bytes
                     0x1017 : "E-SNonce2", # 16 Bytes
                     0x1018 : "Encrypted Settings",
                     0x100F : "Encryption Type", # 2 Bytes
                     0x1010 : "Encryption Type Flags", # 2 Bytes
                     0x101A : "Enrollee Nonce", # 16 Bytes
                     0x101B : "Feature ID", # 4 Bytes
                     0x101C : "Identity", # <= 80 Bytes
                     0x101D : "Identity Proof",
                     0x101E : "Key Wrap Authenticator", # 8 Bytes
                     0x101F : "Key Identifier", # 16 Bytes
                     0x1020 : "MAC Address", # 6 Bytes
                     0x1021 : "Manufacturer", # <= 64 Bytes
                     0x1022 : "Message Type", # 1 Byte
                     0x1023 : "Model Name", # <= 32 Bytes
                     0x1024 : "Model Number", # <= 32 Bytes
                     0x1026 : "Network Index", # 1 Byte
                     0x1027 : "Network Key", # <= 64 Bytes
                     0x1028 : "Network Key Index", # 1 Byte
                     0x1029 : "New Device Name", # <= 32 Bytes
                     0x102A : "New Password", # <= 64 Bytes
                     0x102C : "OOB Device Password", # <= 58
                     0x102D : "OS Version", # 4 Bytes
                     0x102F : "Power Level", # 1 Byte
                     0x1030 : "PSK Current", # 1 Byte
                     0x1031 : "PSK Max", # 1 Byte
                     0x1032 : "Public Key", # 192 Bytes
                     0x1033 : "Radio Enabled", # Bool
                     0x1034 : "Reboot", # Bool
                     0x1035 : "Registrar Current", # 1 Byte
                     0x1036 : "Registrar Established", # Bool
                     0x1037 : "Registrar List", # <= 512 Bytes
                     0x1038 : "Registrar Max", # 1 Byte
                     0x1039 : "Registrar Nonce", # 16 Bytes
                     0x103A : "Request Type", # 1 Byte
                     0x103B : "Response Type", # 1 Byte
                     0x103C : "RF Bands", # 1 Byte
                     0x103D : "R-Hash1", # 32 Bytes
                     0x103E : "R-Hash2", # 32 Bytes
                     0x103F : "R-SNonce1", # 16 Bytes
                     0x1040 : "R-SNonce2", # 16 Bytes
                     0x1041 : "Selected Registrar", # Bool
                     0x1042 : "Serial Number", # <= 32 Bytes
                     0x1044 : "Wi-Fi Protected Setup State", # 1 Byte
                     0x1045 : "SSID", # <= 32 Bytes
                     0x1046 : "Total Networks", # 1 Byte
                     0x1047 : "UUID-E", # 16 Bytes
                     0x1048 : "UUID-R", # 16 Bytes
                     0x1049 : "Vendor Extension", # <= 1024 Bytes
                     0x104A : "Version", # 1 Byte (int)
                     0x104B : "X.509 Certificate Request",
                     0x104C : "X.509 Certificate",
                     0x104D : "EAP Identity", # <= 64 Bytes
                     0x104E : "Message Counter", # 8 Bytes
                     0x104F : "Public Key Hash", # 20 Bytes
                     0x1050 : "Rekey Key", # 32 Bytes
                     0x1051 : "Key Lifetime", # 4 Bytes
                     0x1052 : "Permitted Config Methods", # 2 Bytes
                     0x1053 : "Selected Registrar Config Methods", # 2 Bytes
                     0x1054 : "Primary Device Type", # 8 Bytes
                     0x1055 : "Secondary Device Type List", # <= 128 Bytes
                     0x1056 : "Portable Device", # Bool
                     0x1057 : "AP Setup Locked", # Bool
                     0x1058 : "Application Extension", # <= 512 Bytes
                     0x1059 : "EAP Type", # 8 Bytes
                     0x1060 : "Initialization Vector", # 32 Bytes
                     0x1061 : "Key Provided Automatically", # Bool
                     0x1062 : "802.1X Enabled", # Bool
                     0x1063 : "AppSessionKey", # <= 128 Bytes
                     0x1064 : "WEPTransmitKey" # 1 Byte
                    }

MacAddress = ""
AccessPoints = []

class Packet(object):
  "Packet Class is a wrapper for pcapy packet."
  def __init__(self, pkt):
    self.header       = pkt[0]
    self.data         = pkt[1]
    self.caplen       = self.header.getcaplen()
    self.length       = self.header.getlen()
    self.timestamp    = datetime.datetime.fromtimestamp(self.header.getts()[0])
    self.microseconds = self.header.getts()[1]

  def getHeader(self):
    return self.header

  def getData(self):
    return self.data

  def getCaptureLength(self):
    return self.caplen

  def getLength(self):
    return self.length

  def getTimestamp(self):
    return self.timestamp

  def getMicroseconds(self):
    return self.microseconds

def getAddressFromList(list):
  "Return a string of a MAC address on a python list."
  return ":".join(map(lambda x: "%02X" % x, list))

def getListFromAddress(address):
  "Return a list from a MAC address string."
  return map(lambda x: int(x, 16), address.split(":"))

def isValidMacAddress(address):
  "Return True if it is a valid mac address."
  macAddress = re.compile("^((?:[0-9a-fA-F]{2}[:]){5}[0-9a-fA-F]{2})$")

  if not macAddress.match(address):
    raise Exception("Invalid MAC Address")

  return True

def getVendor(address):
  "Return Vendor string for a MAC address using oui.txt."
  macAddress = re.compile("^((?:[0-9a-fA-F]{2}[:-]){5}[0-9a-fA-F]{2})$")
  ouiLine    = re.compile("^((?:[0-9A-F]{2}[-]){2}[0-9A-F]{2})\s+\(hex\)\s+(.*)$")
  
  if not macAddress.match(address):
    raise Exception("Invalid MAC Address")
  
  address = address.upper()
  if address.find(":") != -1:
    address = "-".join(address.split(":")[:3])
  else:
    address = "-".join(address.split("-")[:3])
  
  fd = open("oui.txt", "r")
  lines = fd.readlines()
  fd.close()
  
  for line in lines:
    match = ouiLine.match(line)
    if match:
      addr, vendor = match.groups()
      if address == addr:
        return vendor
  
  return "<Unknown Vendor>"

def processBeacon(pkt, context):
  "Process 802.11 Beacon Frame for WPS IE."
  packet    = Packet(pkt)
  data      = packet.getData()
  timestamp = packet.getTimestamp()

  try:
    rtDecoder  = RadioTapDecoder()
    radiotap   = rtDecoder.decode(data)
    management = rtDecoder.get_protocol(dot11.Dot11ManagementFrame)
    beacon     = rtDecoder.get_protocol(dot11.Dot11ManagementBeacon)

    bssid  = getAddressFromList(management.get_bssid())
    essid  = beacon.get_ssid()

    if AccessPoints.count(bssid) == 0 and isWPS(beacon.get_vendor_specific()):
      src = getListFromAddress(MacAddress)
      probeRequest = getProbeRequest(src, essid)
      context.send_bytes(probeRequest)

    return None
  except Exception, e:
    return None

def processProbeResponse(pkt):
  "Process 802.11 Probe Response Frame for WPS IE."
  packet    = Packet(pkt)
  data      = packet.getData()
  timestamp = packet.getTimestamp()

  try:
    rtDecoder     = RadioTapDecoder()
    radiotap      = rtDecoder.decode(data)
    management    = rtDecoder.get_protocol(dot11.Dot11ManagementFrame)
    probeResponse = rtDecoder.get_protocol(dot11.Dot11ManagementProbeResponse)

    bssid  = getAddressFromList(management.get_bssid())
    essid  = probeResponse.get_ssid()

    # If null byte in the SSID IE, its cloacked.
    if essid.find("\x00") != -1:
      essid = "<No ssid>"

    if AccessPoints.count(bssid) == 0:
      AccessPoints.append(bssid)
      vendorIEs = probeResponse.get_vendor_specific()
      if isWPS(vendorIEs):
        vendor  = getVendor(bssid) # Get Vendor using OUI.txt
        wpsInfo = parseWPS(vendorIEs)
        return [bssid, essid, vendor, wpsInfo]

    return None
  except Exception, e:
    return None

def isWPS(IEs):
  "Returns True if WPS Information Element is present."
  for element in IEs:
    oui  = element[0]
    data = element[1]
    if data[0] == "\x04": # WPS IE
      return True
    else:
      return False

def parseWPS(IEs):
  "Returns dictionary with WPS Information."
  ret = {}
  for element in IEs:
    offset = 0
    oui  = element[0]
    data = element[1]
    if data[offset] == "\x04": # WPS IE
      offset += 1
      while(offset < len(data)):
        tagType = struct.unpack("!H", data[offset:offset+2])[0]
        offset += 2
        tagLen = struct.unpack("!H", data[offset:offset+2])[0]
        offset += 2
        tagData = data[offset:offset+tagLen]
        offset += tagLen

        # Get the Tag Type
        if WPS_DATA_ELEMENTS.has_key(tagType):
          tagType = WPS_DATA_ELEMENTS[tagType]

        if tagType == "Wi-Fi Protected Setup State":
          if tagData == '\x01':
            tagData = "Not Configured"
          elif tagData == '\x02':
            tagData = "Configured"
          else:
            tagData = 'Reserved'
 
        if tagType == "UUID-E":
          aux = ''
          for c in tagData:
            aux += "%02X" % ord(c)
          tagData = aux

        if tagType == "Response Type":
          if tagData == '\x00':
            tagData = 'Enrollee, Info Only'
          elif tagData == '\x01':
            tagData = 'Enrollee, open 802.1X'
          elif tagData == '\x02':
            tagData = 'Registrar'
          elif tagData == '\x03':
            tagData = 'AP'
          else:
            tagData = '<unkwon>'

        if tagType == "Primary Device Type":
          category    = struct.unpack("!H", tagData[0:2])[0]
          subCategory = struct.unpack("!H", tagData[6:8])[0]
          if category == 1:
            category = "Computer"
            if subCategory == 1:
              subCategory = "PC"
            elif subCategory == 2:
              subCategory = "Server"
            elif subCategory == 3:
              subCategory = "Media Center"
            else:
              subCategory = "<unkwon>"
          elif category == 2:
            category = "Input Device"
            subCategory = "<unkwon>"
          elif category == 3:
            category = "Printers, Scanners, Faxes and Copiers"
            if subCategory == 1:
              subCategory = "Printer"
            elif subCategory == 2:
              subCategory = "Scanner"
            else:
              subCategory = "<unkwon>"
          elif category == 4:
            category = "Camera"
            if subCategory == 1:
              subCategory = "Digital Still Camera"
            else:
              subCategory = "<unkwon>"
          elif category == 5:
            category = "Storage"
            if subCategory == 1:
              subCategory = "NAS"
            else:
              subCategory = "<unkwon>"
          elif category == 6:
            category = "Network Infrastructure"
            if subCategory == 1:
              subCategory = "AP"
            elif subCategory == 2:
              subCategory = "Router"
            elif subCategory == 3:
              subCategory = "Switch"
            else:
              subCategory = "<unkwon>"
          elif category == 7:
            category == "Display"
            if subCategory == 1:
              subCategory = "Television"
            elif subCategory == 2:
              subCategory = "Electronic Picture Frame"
            elif subCategory == 3:
              subCategory = "Projector"
            else:
              subCategory = "<unkwon>"
          elif category == 8:
            category = "Multimedia Devices"
            if subCategory == 1:
              subCategory = "DAR"
            elif subCategory == 2:
              subCategory = "PVR"
            elif subCategory == 3:
              subCategory = "MCX"
            else:
              subCategory = "<unkwon>"
          elif category == 9:
            category = "Gaming Devices"
            if subCategory == 1:
              subCategory = "Xbox"
            elif subCategory == 2:
              subCategory = "Xbox360"
            elif subCategory == 3:
              subCategory = "Playstation"
            else:
              subCategory = "<unkwon>"
          elif category == 10:
            category = "Telephone"
            if subCategory == 1:
              subCategory = "Windows Mobile"
            else:
              subCategory = "<unkwon>"
          else:
            category    = "<unkwon>"
            subCategory = "<unkwon>"
          tagData = "%s - %s" % (category, subCategory)

        if tagType == "Version":
          tagData = struct.unpack("B", tagData)[0]
          major = tagData >> 4
          minor = tagData & 0x0F 
          tagData = "%d.%d" % (major, minor)

        if tagType == "Config Methods":
          methods = {
                    0x0001 : "USB",
                    0x0002 : "Ethernet",
                    0x0004 : "Label",
                    0x0008 : "Display",
                    0x0010 : "External NFC Token",
                    0x0020 : "Integrated NFC Token",
                    0x0040 : "NFC Interface",
                    0x0080 : "PushButton",
                    0x0100 : "Keypad"
                    }
          result = []
          tagData = struct.unpack("!H", tagData)[0]
          for key, value in methods.items():
            if key & tagData:
              result.append(value)
          tagData = ", ".join(result)

        ret[tagType] = tagData
        
  return ret

def getProbeRequest(src, ssid):
  "Return 802.11 Probe Request Frame."

  # Frame Control
  frameControl = Dot11()
  frameControl.set_version(0)
  frameControl.set_type_n_subtype(Dot11Types.DOT11_TYPE_MANAGEMENT_SUBTYPE_PROBE_REQUEST)
  # Frame Control Flags
  frameControl.set_fromDS(0)
  frameControl.set_toDS(0)
  frameControl.set_moreFrag(0)
  frameControl.set_retry(0)
  frameControl.set_powerManagement(0)
  frameControl.set_moreData(0)
  frameControl.set_protectedFrame(0)
  frameControl.set_order(0)

  # Management Frame
  sequence = random.randint(0, 4096)
  broadcast = [0xff, 0xff, 0xff, 0xff, 0xff, 0xff]
  mngtFrame = Dot11ManagementFrame()
  mngtFrame.set_duration(0)
  mngtFrame.set_destination_address(broadcast)
  mngtFrame.set_source_address(src)
  mngtFrame.set_bssid(broadcast)
  mngtFrame.set_fragment_number(0)
  mngtFrame.set_sequence_number(sequence)

  # Probe Request Frame
  probeRequestFrame = Dot11ManagementProbeRequest()
  probeRequestFrame.set_ssid(ssid)
  probeRequestFrame.set_supported_rates([0x02, 0x04, 0x0b, 0x16])

  # How is your daddy?
  mngtFrame.contains(probeRequestFrame)
  frameControl.contains(mngtFrame)
 
  return frameControl.get_packet()

def processPacket(pkt, context):
  "Process 802.11 Packets."

  packet = Packet(pkt)
  data = packet.getData()

  try:
    rtDecoder     = RadioTapDecoder()
    radiotap      = rtDecoder.decode(data)
    management    = rtDecoder.get_protocol(dot11.Dot11ManagementFrame)
    beacon        = rtDecoder.get_protocol(dot11.Dot11ManagementBeacon)
    probeResponse = rtDecoder.get_protocol(dot11.Dot11ManagementProbeResponse)

    if beacon:
      processBeacon(pkt, context)

    if probeResponse:
      info = processProbeResponse(pkt)
      if info:
        bssid   = info[0]
        essid   = info[1]
        vendor  = info[2]
        wpsInfo = info[3]
        result  = "[%s] - '%s' - '%s'\n" % (bssid, essid, vendor)
        result += "WPS Information\n"
        for key, value in wpsInfo.items():
          result += "  * %s: %s\n" % (key, repr(value))
        result += "-" * 80
        result += "\n"
      return result

  except Exception:
    return None

def doYourThing(pd, context, fd):
  try:
    packet = pd.next()
  except pcapy.PcapError:
    packet = None
  while(True):
    if packet:
      data = processPacket(packet, context)
      if data:
        print data,
        if fd:
          fd.write(data)
    try:
      packet = pd.next()
    except pcapy.PcapError:
      packet = None

if __name__ == "__main__":
  # Allways print the title.
  print "Wi-Fi Protected Setup Information Gathering."

  usage = "%prog -i interface -w filename"

  parser = optparse.OptionParser(usage)
  parser.add_option("-i", "--interface", dest="iface",
                    type="string", help="network interface where to sniff.")
  parser.add_option("-w", "--write", dest="filename",
                    type="string", help="output filename.")
  parser.add_option("-s", "--source", dest="source",
                    type="string", help="source mac address of the probe request frames.")

  options, args = parser.parse_args()

  # Mandatory Option
  if not options.__dict__["iface"]:
    parser.print_help()
    sys.exit(-1)

  if options.__dict__["source"]:
    if isValidMacAddress(options.source):
      MacAddress = options.source
  else:
      MacAddress = "00:00:00:11:22:33" # Default Mac Address for injection

  interface = options.iface
  filename  = options.filename

  try:
    # Setup Interface
    context = PyLorcon2.Context(interface)
    context.open_injmon()
    moniface = context.get_capiface()

    pd = pcapy.open_live(moniface, 65535, 0, 0)

    # Show Message
    print "Press Ctrl+C to stop."
    print "Sniffing..."
    print "-" * 80

    if filename:
      fd = open(filename, "w")
    else:
      fd = None

    doYourThing(pd, context, fd)
  except KeyboardInterrupt:
    if fd:
      fd.close()
    print "\nCtrl+C caught."
    print "Closing..."
